from random import random, randrange

from .config import cfg
from .session import calc_progress


class WordIterator:
    def __init__(self, words: list):
        self.__words = words
        self.__add_after_iter = None

    def __iter__(self):
        return self

    def add_word(self, index: int, value):
        self.__add_after_iter = (index, value)

    def __next__(self):
        ws = self.__words
        if not any(ws[:-1]):
            i, res = self.__add_after_iter
            self.__add_after_iter = None
            if i < len(self.__words):
                return i == 1, i, res
            raise StopIteration

        # Get non-empty buckets, position new word bucket at nwi
        nwi = cfg['new-words-index']
        ws = [l for l in ws[1:nwi+1] + [ws[0]] + ws[nwi+1:] if l]

        # Pick a random word, favor lower indices
        l = ws[int(random() ** cfg['critical-word-weight'] * len(ws))]
        img_idx = randrange(len(l))
        i = next(i for i, l_ in enumerate(self.__words) if l_ is l)

        if self.__add_after_iter:
            index, value = self.__add_after_iter
            self.__add_after_iter = None
            self.__words[min(len(self.__words)-1, index)].append(value)

        # Calculate progress before popping the word
        return i == 0, i, calc_progress(self.__words), l.pop(img_idx)



import ctypes
import json
import os

import PySimpleGUI as sg
from pkg_resources import resource_filename


def data_path(*args):
    return os.path.join(os.getenv('APPDATA'), 'pupquiz', *args)


cfg = {
    # Regex patterns
    'patt-vocab-config': r'^\s*<!--(.+)-->\r?\n',
    'patt-set': r'(?i)^(?P<set>.+\\)(?P<img>.+\.(jpg|gif|png))$',
    'patt-gif-loop-pause': r'(?i)(?<!p)\.gif$',
    'patt-vocab-word': r'(?m)^(?P<q>.+)\n[^\S\n]*~ (?P<a>.+)',
    'patt-vocab-word-subdivider': r'<br>|<i>.*?</i>',
    'patt-word-spoken-part': r'^[^,]+',
    'patt-vocab-file': r"""(?ms)^title: "(?P<title>.+?)"$.+?^content: '''(?P<contents>.+?^)'''""",

    # Selection screen
    'select-hotkeys': '123456789',

    # Vocabularies
    'vocab-file-types': [['Vocabulary Files', '*.cson']],

    # Sets
    'common-sets': [data_path('Default Assets')],
    'gif-pause-duration': 1.2,

    # Session
    'spoken-lang': '',
    'new-words-index': 2,
    'translate-url': 'https://translate.google.com/#auto/en/{}',
    'critical-word-weight': 3.5,

    # Locale
    'locale': ['Pup Quiz', 'Click on a vocabulary to quiz for it.\nAdd a vocabulary by selecting an empty slot.\nDelete a vocabulary with a right-click.', '{} - Pup Quiz', 'Welcome to Pup Quiz!', 'Correct!', '{}', 'Added to word stock', 'Got it!', 'Submit', 'Translate', 'Translation opened', 'Reset', 'Menu', 'Are you sure you want to delete "{}"?', 'Are you sure you want to reset progress on "{}"?'],

    # Styling
    'app-icon': resource_filename('pupquiz.resources', 'icon.ico'),
    'image-folder-on': resource_filename('pupquiz.resources', 'folder-on.png'),
    'image-folder-off': resource_filename('pupquiz.resources', 'folder-off.png'),
    'color-text': '#f0eff4',
    'color-select-info-guide': '#f0eff4',
    'color-background': '#111111',
    'color-input-text': '#f0eff4',
    'color-input-background': '#3d2645',
    'color-button-text': '#f0eff4',
    'color-button-background': '#832161',
    'color-info-incorrect': '#da4167',
    'color-info-correct': '#f0eff4',
    'color-info-greet': '#f0eff4',
    'color-info-new-word': '#f0eff4',
    'color-info-translation-opened': '#f0eff4',
    'select-info-guide-width': 4,
    'select-info-guide-height': 59,
    'pad-top-status': 150,
    'image-max-width': 400,
    'image-max-height': 500,
    'control-border-width': 0,
    'font': '"Berlin Sans FB" 13',
    'thumbnail-placeholder-color-hover': '#626262',
    'thumbnail-placeholder-color-default': '#454545',
    'thumbnail-placeholder-width': 10,
    'thumbnail-size': 128,
    'thumbnail-sample': 4,
    'thumbnail-font-file': 'BRLNSR.TTF',
    'thumbnail-font-size': 16,
    'thumbnail-font-color': '#f0eff4',
    'thumbnail-text-stroke-color': '#000000',
    'thumbnail-text-stroke-size': 1,
    'thumbnail-border-color': '#333333',
    'thumbnail-border-size': 1,
    'thumbnail-undone-brightness': 0.75,
    'thumbnail-undone-saturation': 0.4,
    'thumbnail-undone-blur-radius': 3,
    'background-brightness': 0.6,
    'background-saturation': 0.8,
    'background-blur-radius': 10,
}


CFG_PATH = data_path('config-user.json')


if not os.path.isdir(data_path()):
    os.mkdir(data_path())

if not os.path.isdir(data_path('Default Assets')):
    from zipfile import ZipFile
    from pkg_resources import resource_stream
    with ZipFile(resource_stream('pupquiz.resources', 'default-assets.zip'), 'r') as f:
        f.extractall(data_path())


if not os.path.isfile(data_path('config-default.json')):
    with open(data_path('config-default.json'), 'w') as f:
        json.dump(cfg, f, indent=4)  # for reference
try:
    with open(CFG_PATH, 'r') as f:
        cfg.update(json.load(f) or {})  # apply user preference
except FileNotFoundError:
    with open(CFG_PATH, 'w') as f:
        f.write('{\n}')


# Set icon
# https://stackoverflow.com/a/34547834
myappid = 'mycompany.myproduct.subproduct.version'
ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
sg.set_global_icon(cfg['app-icon'])

# Set GUI theme
sg.theme_background_color(cfg['color-background'])
sg.theme_element_background_color(cfg['color-background'])
sg.theme_text_color(cfg['color-text'])
sg.theme_text_element_background_color(cfg['color-background'])
sg.theme_input_text_color(cfg['color-input-text'])
sg.theme_input_background_color(cfg['color-input-background'])
sg.theme_button_color(
    (cfg['color-button-text'], cfg['color-button-background']))
sg.theme_border_width(cfg['control-border-width'])

# Set locale
CFG_APPNAME, CFG_SELECT_INFO, CFG_APPNAME_SES, CFG_GREET, CFG_CORRECT, CFG_INCORRECT, CFG_ADDWORD, CFG_NEWWORD, CFG_GUESS, CFG_TRANSLATE, CFG_TRANSLATE_OPENED, CFG_RESET, CFG_MENU, CFG_CONFIRM_DELETE, CFG_CONFIRM_RESET = cfg[
    'locale']



import gc
import os
import re
import webbrowser
from contextlib import suppress
from threading import Lock, Thread

from gtts import gTTS
from playsound import playsound

from .canvas import CANVAS_SZ, Canvas
from .config import *
from .session import (SES_WIN_POS, VOCAB_CONFIG, VOCAB_NAME, VOCAB_WORDS, Set,
                      SetProvider, calc_progress, get_vocabulary,
                      save_session, ses, sg, update_vocab)
from .word_iterator import WordIterator


def weighted_pick(ls, rd):
    return ls[int(rd.random()**3.5*len(ls))]


FORM_WIDTH = 28
TTS_PATH = data_path('tts.mp3')


def speak_tts(word: str):
    if os.path.exists(TTS_PATH):
        os.remove(TTS_PATH)
    try:
        gTTS(word, lang=cfg['spoken-lang']).save(TTS_PATH)
        playsound(TTS_PATH)
    except:
        pass


class Quiz:
    def __init__(self, v: dict, sets: SetProvider):
        self.__v = v
        self.__sets = sets
        self.__words = v[VOCAB_WORDS]
        self.__nwords = sum(map(len, self.__words))
        self.__nsteps = self.__nwords * (len(self.__words)-2)

    def run(self):
        # Create command card layout
        def buts(*args):
            return [sg.B(text, bind_return_key=key == '-OK-', key=key, pad=((0, 7), (10, 0))) for text, key in args]
        cc_layout = [[sg.T(CFG_GREET, pad=(0, 0), key='-RES-', size=(FORM_WIDTH, None))], [sg.T(key='-TXT-', pad=(0, (30, 10)), size=(FORM_WIDTH, None))]
                     ] + [[sg.In(key=0, focus=True, pad=(0, 0), size=(FORM_WIDTH, 100))]] + [buts(('', '-OK-'), (CFG_TRANSLATE, '-TRANSL-'), (CFG_RESET, '-RESET-'), (CFG_MENU, '-MENU-'))]

        # Full layout including image tiles
        layout = [[sg.Image(size=CANVAS_SZ, key='-IM1-'), sg.Column([[sg.Image(size=(0, 0), key='-IM2-')], [sg.Col(cc_layout, key='-CCARD-', pad=(10, 10), size=(300, 200))], [
            sg.Image(size=(0, 0), key='-IM3-')]])]]

        # Create window
        win = sg.Window(CFG_APPNAME_SES.format(self.__v[VOCAB_NAME]), layout, location=ses[SES_WIN_POS],
                        finalize=True, font=cfg['font'], border_depth=0, margins=(0, 0), element_padding=(0, 0))
        win.hide()
        hidden = True
        canvas = Canvas(win)

        no_advance = False
        it = WordIterator(self.__words)
        for new, bucket, progress, word in it:
            print(f'{progress}')
            win['-OK-'].update(CFG_NEWWORD if new else CFG_GUESS)

            # Set image
            set_, img = self.__sets.get_image(progress)
            canvas.set_image(set_, img, no_advance)
            no_advance = True

            # Speak new words
            if new and cfg['spoken-lang']:
                if m := re.search(cfg['patt-word-spoken-part'], word[0]):
                    Thread(target=speak_tts, args=(m[0],), daemon=True).start()

            # Reset controls
            win['-TRANSL-'].update(disabled=len(word) != 2)
            win['-TXT-'].update(word[-1])
            col, bgcol = (cfg['color-text'], cfg['color-background']) if new else (
                cfg['color-input-text'], cfg['color-input-background'])
            for i in range(1):
                win[i].update(word[i] if new else '',
                              text_color=col, background_color=bgcol, select=not new)
            win[0].set_focus()

            # Unhide window, if hidden
            if hidden:
                win.un_hide()
                win.TKroot.focus_force()
                hidden = False

            # Retrieve input
            while True:
                event, values = win.read(10)
                if event != sg.TIMEOUT_KEY:
                    if event == '-RESET-':
                        if not self.__sets.reset_progress():
                            continue
                    break
                ses[SES_WIN_POS] = list(win.CurrentLocation())
                canvas.update()
            if event in (None, '-MENU-'):
                self.__words[bucket].append(word)
                break

            # 'Translate' button
            elif event == '-TRANSL-':
                self.__words[bucket].append(word)
                win['-RES-'].update(CFG_TRANSLATE_OPENED,
                                    text_color=cfg['color-info-translation-opened'])
                webbrowser.open(cfg['translate-url'].format(word[0]))

            elif event == '-OK-':

                # 'Got it!' button
                if new:
                    it.add_word(1, word)
                    win['-RES-'].update(CFG_ADDWORD,
                                        text_color=cfg['color-info-new-word'])

                # 'Submit' button
                else:
                    # Compare answer with solution
                    guess = list(values.values())
                    if all(map(lambda x: x[0] in [y.strip() for y in x[1].split(',')], zip(guess, word[:-1]))):
                        it.add_word(bucket+1, word)
                        win['-RES-'].update(CFG_CORRECT,
                                            text_color=cfg['color-info-correct'])
                        no_advance = False
                    else:
                        it.add_word(1, word)
                        win['-RES-'].update(CFG_INCORRECT.format(
                            ', '.join(word[:-1])), text_color=cfg['color-info-incorrect'])

        # Submit progress to disk and quit
        win.close()
        layout = None
        cc_layout = None
        win = None
        gc.collect()
        update_vocab(self.__v)
        return type(event) == str and event == '-MENU-'


def main_loop():
    while True:
        v, provider = get_vocabulary()
        if not v:
            break

        # Support overriding config on a per-vocab basis
        vconfig = {**v[VOCAB_CONFIG]}
        for k in vconfig:
            original = cfg[k]
            cfg[k] = vconfig[k]
            vconfig[k] = original

        if not Quiz(v, provider).run():
            break

        # Restore common config
        cfg.update(vconfig)

    save_session()
